@@ -9,40 +9,27 @@ Distribute jobs over multiple machines by means of ssh.
 - restart failed
 """
 
+from argparse import ArgumentParser, SUPPRESS
+from collections import defaultdict, OrderedDict
+from datetime import datetime
 from fnmatch import fnmatch
+from functools import partial
 from logging import warning
+from math import ceil
+from os import remove
+from os.path import basename, join, exists, isfile
+from random import sample
 from subprocess import PIPE
 from subprocess import Popen
 from sys import stdout, stderr
 from time import time, sleep
-from random import sample
 from bardeen.inout import reprint
-from datetime import datetime
-from collections import defaultdict, OrderedDict
-from argparse import ArgumentParser, SUPPRESS
-from os import remove
-from os.path import basename, join
-from math import ceil
-from functools import partial
-from .shell import run_cmds_on, run_cmds
+from fenpei.utils import job_task, compare_jobs, compare_results, job_results
 from .job import Job
+from .shell import run_cmds_on
 from .utils import get_pool_light, TMP_DIR, thread_map
 
 
-def job_task_run(job, method, **kwargs):
-	"""
-	Runs an arbitrary method of job; used by job_task.
-	"""
-	return getattr(job, method)(**kwargs)
-
-
-def job_task(method, **kwargs):
-	"""
-	Returns a function that runs an arbitrary method of an object, for passing to Pool.map
-	"""
-	return partial(job_task_run, method=method, **kwargs)
-
-
 class Queue(object):
 
 	def __init__(self, jobs=None, summary_func=None):
@@ -80,7 +67,7 @@ class Queue(object):
 		self._log('finding nodes')
 		self.nodes = []
 		self.slots = []
-		''' find node ssh adresses and store in self.nodes '''
+		""" find node ssh adresses and store in self.nodes """
 		return True
 
 	def node_availability(self):
@@ -90,23 +77,23 @@ class Queue(object):
 		if self.load_nodes():
 			return False
 		if not len(self.nodes):
-			self._log('no nodes yet; calling all_nodes()', level = 2)
+			self._log('no nodes yet; calling all_nodes()', level=2)
 			self.all_nodes()
 			if not len(self.nodes):
-				self._log('no nodes found; no availability checked', level = 2)
+				self._log('no nodes found; no availability checked', level=2)
 				return
 		self.slots = []
-		self._log('checking node availability', level = 1)
+		self._log('checking node availability', level=1)
 		for node in self.nodes:
 			outps = run_cmds_on(cmds = ['grep \'model name\' /proc/cpuinfo | wc -l', 'uptime'], node = node, queue = self)
 			if len(outps) == 2:
-				''' one slot for every 100% processor available '''
+				""" one slot for every 100% processor available """
 				proc_count = int(outps[0])
 				load_1min = float(outps[1].split()[-3].replace(',', ''))
 				self.slots.append(max(proc_count - load_1min, 0))
-				self._log('%2d slots assigned to %6s - 1min cpu %4d%% on %d processors' % (round(self.slots[-1]), self.short_node_name(node), 100 * load_1min, proc_count), level = 2)
+				self._log('%2d slots assigned to %6s - 1min cpu %4d%% on %d processors' % (round(self.slots[-1]), self.short_node_name(node), 100 * load_1min, proc_count), level=2)
 			else:
-				''' not accessible for some reason '''
+				""" not accessible for some reason """
 				self._log('%s not accessible' % node)
 				self.nodes.remove(node)
 		self._log('found %d idle processors on %d nodes' % (sum(self.slots), len(self.nodes)))
@@ -172,19 +159,19 @@ class Queue(object):
 		if jobs is None:
 			jobs = self.jobs
 		assert len(self.nodes) == len(self.slots)
-		max_reject_spree = 2 * len(self.nodes) if max_reject_spree is None else max_reject_spree
+		# max_reject_spree = 2 * len(self.nodes) if max_reject_spree is None else max_reject_spree
 		self._log('distributing %d jobs with weight %d over %d slots' % (len(jobs), self.total_weight(jobs), sum(self.slots)))
 		def cost(weight_1, slots_1, weight_2, slots_2):
 			return max(weight_1 - slots_1, 0) ** 2 + max(weight_2 - slots_2, 0) ** 2 + slots_1 / max(weight_1, 1) + slots_2 / max(weight_2, 1)
-		''' clear the list '''
+		""" clear the list """
 		distribution = {}
 		for node_nr in range(len(self.nodes)):
 			distribution[node_nr] = []
-		''' random initial job distribution '''
+		""" random initial job distribution """
 		for job in jobs:
 			node_nr = sample(distribution.keys(), 1)[0]
 			distribution[node_nr].append(job)
-		''' repeat switching until nothing favourable is found anymore '''
+		""" repeat switching until nothing favourable is found anymore """
 		reject_spree, steps = 0, 0
 		while reject_spree < 100:
 			node1, node2 = sample(distribution.keys(), 2)
@@ -195,36 +182,37 @@ class Queue(object):
 				item1 = sample(range(len(distribution[node1])), 1)[0]
 				cost_switch = cost_move = None
 				if len(distribution[node2]) > 0:
-					''' compare the cost of switching two items '''
+					""" compare the cost of switching two items """
 					item2 = sample(range(len(distribution[node2])), 1)[0]
 					cost_switch = cost(self.total_weight(distribution[node1]) - distribution[node1][item1].weight + distribution[node2][item2].weight, self.slots[node1],
 									   self.total_weight(distribution[node2]) + distribution[node1][item1].weight - distribution[node2][item2].weight, self.slots[node2])
 				if cost_before > 0:
-					''' compare the cost of moving an item '''
+					""" compare the cost of moving an item """
 					cost_move = cost(self.total_weight(distribution[node1]) - distribution[node1][item1].weight, self.slots[node1],
 									 self.total_weight(distribution[node2]) + distribution[node1][item1].weight, self.slots[node2])
-				''' note that None < X for any X, so this works even if only cost_before has an actual value '''
+				""" note that None < X for any X, so this works even if only cost_before has an actual value """
 				if (cost_switch < cost_before and cost_switch is not None) or (cost_move < cost_before and cost_move is not None):
 					if cost_switch < cost_move and cost_switch is not None:
-						''' switch '''
+						""" switch """
 						tmp = distribution[node1][item1]
 						distribution[node1][item1] = distribution[node2][item2]
 						distribution[node2][item2] = tmp
 					elif cost_move is not None:
-						''' move (move if equal, it's easier after all) '''
+						""" move (move if equal, it's easier after all) """
 						distribution[node2].append(distribution[node1][item1])
 						del distribution[node1][item1]
 					reject_spree = 0
 				else:
-					''' not favorable; don't move '''
+					""" not favorable; don't move """
 					reject_spree += 1
 			else:
-				''' too many empty slots means few rejectsbut lots of iterations, so in itself a sign to stop '''
+				""" too many empty slots means few rejectsbut lots of iterations, so in itself a sign to stop """
 				reject_spree += 0.1
 		self.distribution = distribution
-		''' report results '''
-		self._log('distribution found after %d steps' % steps)
-		self._log(self.text_distribution(distribution), level = 2)
+		""" report results """
+		self._log('distribution found after {0:d} steps'.format(steps))
+		self._log(self.text_distribution(distribution), level=2)
+		return self.distribution
 
 	def text_distribution(self, distribution):
 		"""
@@ -277,7 +265,7 @@ class Queue(object):
 		if node in self.process_time.keys():
 			if time() - self.process_time[node] < 3:
 				return self.process_list[node]
-		self._log('loading processes for %s' % node, level = 3)
+		self._log('loading processes for %s' % node, level=3)
 		self.process_time[node] = time()
 		self.process_list[node] = []
 		outp = run_cmds_on([
@@ -321,6 +309,18 @@ class Queue(object):
 	def get_jobs(self):
 		return self.jobs
 
+	def compare_jobs(self, parameters, filter=None):
+		return compare_jobs(self.jobs, parameters, filter=filter)
+	
+	def compare_results(self, parameters, filter=None):
+		return compare_results(self.jobs, parameters, filter=filter)
+	
+	def result(self, parallel=None, *args, **kwargs):
+		parallel = self.parallel if parallel is None else parallel
+		results = job_results(parallel=parallel, *args, **kwargs)
+		self._log('retrieved results for %d jobs' % len(self.jobs))
+		return results
+
 	def list_jobs(self, cols=2, verbosity=0, *args, **kwargs):
 		N = int(ceil(len(self.jobs) / float(cols)))
 		for k in range(N):
@@ -332,6 +332,21 @@ class Queue(object):
 					for p in [k, k+N, k+2*N] if p < len(self.jobs)
 			) + '\n')
 
+	def json_jobs(self, *args, **kwargs):
+		maxbatchlen = max(len(job.batch_name or '') for job in self.jobs) + 2
+		maxjoblen = max(len(job.name) for job in self.jobs) + 2
+		stdout.write('[\n')
+		for k, job in enumerate(sorted(self.jobs, key=lambda job: job.name)):
+			stdout.write('  {{"batch": {batch:s}, "name": {name:s}, "weight": {weight:6d}, "status": {status:10s}}}'.format(
+				batch=('"' + (job.batch_name or '') + '"').ljust(maxbatchlen),
+				name=('"' + job.name + '"').ljust(maxjoblen),
+				weight=job.weight,
+				status='"' + job.status_str() + '"',
+			))
+			if k + 1 < len(self.jobs):
+				stdout.write(',\n')
+		stdout.write('\n]\n')
+
 	def run_job(self, job, filepath):
 		"""
 		Start an individual job, specified by a Python file.
@@ -375,72 +390,89 @@ class Queue(object):
 		"""
 		parallel = self.parallel if parallel is None else parallel
 		if parallel:
-			#statuses = get_pool_light().map(job_task('prepare', **kwargs), self.jobs)
 			statuses = thread_map(job_task('prepare', **kwargs), self.jobs)
 		else:
 			statuses = (job.prepare(**kwargs) for job in self.jobs)
 		prepare_count = sum(int(status) for status in statuses)
 		self._log('prepared %d jobs' % prepare_count)
 
-	def running_count(self):
-		"""
-		How many running jobs.
-		"""
-		return len(self.get_status()[Job.RUNNING])
-
-	def running_weight(self):
-		"""
-		Total weight of running jobs.
-		"""
-		return sum(job.weight for job in self.get_status()[Job.RUNNING])
-
 	def start(self, parallel=None, verbosity=0, *args, **kwargs):
 		"""
 		Calls corresponding functions depending on flags (e.g. -z, -w, -q, -e).
 		"""
 		self._quota_warning()
 		self._same_path_check(fail=True)
-		# if self.all:
-		# 	if self.weight:
-		# 		self._log('starting all jobs; specific weight (-w) ignored')
-		# 	if self.limit:
-		# 		self._log('starting all jobs; limit weight (-q) ignored')
-		W = float('inf')
-		if self.limit:
-			W = max(self.limit - self.running_weight(), 0)
-			if verbosity:
-				if not self.weight:
-					self._log('starting jobs with weight %d (no minimum)' % W)
-				elif W < self.weight:
-					self._log('starting jobs with weight %d because of minimum weight %d' % (W, self.weight))
-				else:
-					self._log('starting jobs with weight %d to fill to %d (higher than minimum)' % (W, self.limit))
-		elif self.weight:
-			W = self.weight
-			self._log('starting jobs with weight %d (by fixed weight)' % self.weight)
-		self.start_weight(W, parallel=parallel)
+		job_status = self.get_status()
+		start_jobs = self.select_start_jobs(weight=self.weight, limit=self.limit,
+			restart=self.restart, job_status=job_status)
+		self._log('starting {0:d} jobs with weight {1:d}'.format(
+			len(start_jobs), sum((job.weight for job in start_jobs), 0)), level=2)
+		distribution = self.distribute_jobs(jobs=start_jobs)
+		start_cnt = 0
+		for node_nr, jobs in distribution.items():
+			""" Removed parallism here because it made things slower (and more complicated). """
+			for job in jobs:
+				node = self.nodes[node_nr]
+				if job.is_started():
+					job.cleanup(**kwargs)
+				start_cnt += job.start(node, **kwargs)
+		self._log('started {0:d} jobs'.format(start_cnt, level=1))
 
-	def start_weight(self, weight, parallel=None, *args, **kwargs):
+	def select_start_jobs(self, weight, limit, restart, job_status=None):
 		"""
-		(Re)start jobs with an approximation of total weight.
+		Find jobs in job_status or self.get_status() that can be started to stay within the weight and count limits provided.
+		
+		:param weight: Total running weight limit.
+		:param limit: Total running count limit.
+		:param restart: Whether crashed jobs should be restarted (at comparable weight, start fresh ones first).
+		:return:
 		"""
-		parallel = self.parallel if parallel is None else parallel
-		parallel = False  # override because it's slow in this case, somehow...
-		batch = []
-		jobs = self.get_jobs_by_weight(weight)
-		if len(jobs):
-			self.distribute_jobs(jobs = jobs)
-			for node_nr, jobs in self.distribution.items():
-				for job in jobs:
-					batch.append((job, self.nodes[node_nr]))
-			if parallel:
-				statuses = get_pool_light().map(partial(_start_job_on_node, **kwargs), batch)
-			else:
-				statuses = map(partial(_start_job_on_node, **kwargs), batch)
-			start_count = sum(int(status) for status in statuses)
-			self._log('(re)started %d jobs' % start_count if self.restart else 'started %d jobs' % start_count)
+		""" Find jobs with startable status. """
+		if job_status is None:
+			job_status = self.get_status()
+		startable = job_status[Job.PREPARED] + job_status[Job.NONE]
+		for job in startable:
+			job._crash_score = 1
+		if restart:
+			self._log('crashed jobs are eligible for restarting', level=2)
+			for job in job_status[Job.CRASHED]:
+				job._crash_score = 0
+			startable.extend(job_status[Job.CRASHED])
 		else:
-			self._log('no jobs to restart' if self.restart else 'no jobs to start')
+			self._log('only unstarted jobs will be started', level=2)
+		if not self.weight and not self.limit:
+			return startable
+		startable = sorted(startable, key=lambda job: job.weight + 2 * job._crash_score, reverse=False)
+		""" Remove jobs to reach the requested weight. """
+		if weight is not None:
+			start_jobs = []
+			running_weight = sum(job.weight for job in job_status[Job.RUNNING])
+			self._log('{0:d} jobs with weight {1:d} already running'.format(len(job_status[Job.RUNNING]), running_weight), level=2)
+			weight_left = weight - running_weight
+			while startable and weight_left > 0:
+				consider_job = startable.pop()
+				if weight_left > consider_job.weight:
+					weight_left -= consider_job.weight
+					start_jobs.append(consider_job)
+			self._log('pre-selecting {0:d} jobs with weight {1:d} to stay under the weight limit of {2:d}'
+				.format(len(start_jobs), sum((job.weight for job in start_jobs), 0), weight), level=2)
+		else:
+			start_jobs = startable
+			self._log('pre-selecting all {0:d} eligible jobs for starting'.format(len(start_jobs)), level=2)
+		""" Limit the total number of jobs to start. """
+		if limit is not None:
+			running_count = len(job_status[Job.RUNNING])
+			start_count = max(min(limit - running_count, len(start_jobs)), 0)
+			self._log(('starting {0:d} of {1:d} eligible jobs, to stay under the limit of {2:d} '
+				'jobs, with {3:d} already running').format(start_count, len(start_jobs), limit,
+				running_count), level=2)
+			start_jobs = start_jobs[-start_count:]
+			if start_count == 0:
+				start_jobs = []
+			assert len(start_jobs) <= start_count, 'too many jobs: {0:d} <= {1:d}'.format(len(start_jobs), start_count)
+		else:
+			self._log('keeping {0:d} for starting since no limit was provided'.format(len(start_jobs)), level=2)
+		return start_jobs
 
 	def _quota_warning(self):
 		try:
@@ -459,7 +491,11 @@ class Queue(object):
 		found = 0
 		for job in self.jobs:
 			if (job.batch_name, job.name) in pths:
-				msg = 'there are multiple jobs with location {0:s}/{1:s}'.format(job.batch_name, job.name)
+				if job.batch_name:
+					pthname = '{0:s}/{1:s}'.format(job.batch_name, job.name)
+				else:
+					pthname = job.name
+				msg = 'there are multiple jobs with location {0:s}'.format(pthname)
 				if fail:
 					raise AssertionError(msg)
 				found += 1
@@ -470,37 +506,6 @@ class Queue(object):
 			pths.add((job.batch_name, job.name))
 		return pths
 
-	def get_jobs_by_weight(self, max_weight):
-		"""
-		Find jobs with an approximation of total weight.
-		"""
-		""" find eligible jobs (in specific order) """
-		job_status = self.get_status()
-		if self.restart:
-			startable = job_status[Job.PREPARED] + job_status[Job.NONE] + job_status[Job.CRASHED]
-		else:
-			startable = job_status[Job.PREPARED] + job_status[Job.NONE]
-		if not startable:
-			self._log('there are no jobs that can be started')
-			return []
-		total_weight = sum(job.weight for job in startable)
-		if not total_weight:
-			""" start only one job """
-			jobs = [startable[0]]
-		elif max_weight > total_weight or max_weight is None:
-			""" start all jobs """
-			jobs = startable
-		else:
-			""" start jobs with specific weight """
-			jobs, current_weight = [], 0
-			startable = sorted(startable, key = lambda item: - item.weight - (10 if item.status == Job.CRASHED else 0))
-			for job in startable:
-				if current_weight + job.weight <= max_weight:
-					jobs.append(job)
-					current_weight += job.weight
-		self._log('starting: ' + ', '.join(job.name for job in jobs), level = 2)
-		return jobs
-
 	def fix(self, parallel=None, *args, **kwargs):
 		"""
 		Fix jobs, e.g. after fixes and updates.
@@ -558,7 +563,7 @@ class Queue(object):
 		"""
 		Show list of statusses.
 		"""
-		self._log('status for %d jobs:' % len(self.jobs), level = 1)
+		self._log('status for %d jobs:' % len(self.jobs), level=1)
 		for status_nr in status_list.keys():
 			if verbosity <= 0:
 				job_names = ' '.join(str(job) for job in status_list[status_nr][:20 ])
@@ -583,13 +588,14 @@ class Queue(object):
 		while True:
 			try:
 				status_list = self.get_status()
-
-				txt = '%s - job# %d; weight %d:' % (datetime.now().strftime('%H:%M:%S'), self.running_count(), self.running_weight())
+				running_count = len(status_list[Job.RUNNING])
+				running_weight = sum((job.weight for job in status_list[Job.RUNNING]), 0)
+				txt = '{0:s} - job# {1:d}; weight {2:d}:'.format(datetime.now().strftime('%H:%M:%S'),
+					running_count, running_weight)
 				for status_nr in status_list.keys():
 					job_names = ', '.join(str(job) for job in status_list[status_nr])
 					txt += '\n %3d %-12s %s' % (len(status_list[status_nr]), Job.status_names[status_nr], job_names if len(job_names) <= 40 else job_names[:37] + '...')
 				reprint(txt, lines)
-
 				if not status_list[Job.RUNNING]:
 					nothing_running_count += 1
 				else:
@@ -613,71 +619,18 @@ class Queue(object):
 		status_list = self.get_status()
 		self.show_status(status_list, verbosity=verbosity)
 
-	def result(self, parallel=None, jobs=None, *args, **kwargs):
-		"""
-		:return: a dict of job results, with jobs as keys.
-		"""
-		parallel = self.parallel if parallel is None else parallel
-		# parallel = False  # override because parallel is much slower for some reason.
-		if jobs is None:
-			jobs = self.jobs
-		results = OrderedDict()
-		if parallel:
-			resli = get_pool_light().map(job_task('result', **kwargs), jobs)
-			for job, res in zip(jobs, resli):
-				results[job] = res
-		else:
-			for job in jobs:
-				results[job] = job.result(*args, **kwargs)
-		for job, res in results.items():
-			if res and 'in' not in res:
-				res['in'] = job.get_input()
-		self._log('retrieved results for %d jobs' % len(jobs))
-		return results
-
+	# @staticmethod
+	# def pre_summary(queue, *args, **kwargs):
+	# 	pass
+	
 	@staticmethod
 	def summary(queue):
-		raise NotImplementedError(('No summary function (queue "{0:s}"). Attach a '
-			'static method .summary(queue) to the queue.').format(queue.name))
+		raise NotImplementedError(('No summary function (queue "{0:}"). Attach a '
+			'static method .summary(queue) to the queue.').format(queue))
 
-	def compare_jobs(self, parameters, filter=None):
-		"""
-		Get a parameters -> job mapping. The parameters are expected to identify unique jobs.
-
-		:param filter: a function that returns True for jobs that should be included.
-		:return: Without parameters, a list of jobs. With parameters, a mapping from parameter to accompanying jobs. Indices are tuples of parameter values.
-		"""
-		if not hasattr(parameters, '__iter__'):
-			parameters = (parameters,)
-		assert len(parameters) > 0, 'Provide a job attribute to compare jobs.'
-		def get_key(j):
-			vals = []
-			for param in parameters:
-				assert hasattr(job, param), 'Can not compare jobs on "{0:s}" since job "{1:s}" does not have this attribute.'.format(param, job)
-				vals.append(getattr(job, param))
-			# if len(vals) == 1:
-			# 	return vals[0]
-			return tuple(vals)
-		jobmap = OrderedDict()
-		if filter is None:
-			filter = lambda obj: True
-		for job in self.jobs:
-			if filter(job):
-				key = get_key(job)
-				assert key not in jobmap, 'Can not compare jobs on "{0:}" since jobs "{1:s}" and "{2:s}" both have value <{3:}>, but values should be unique.'.format(parameters, jobmap[key], job, key)
-				jobmap[key] = job
-		return jobmap
-
-	def compare_results(self, parameters, filter=None):
-		"""
-		Similar to compare_jobs but uses a map from parameters -> results instead. Furthermore, jobs without results are omitted.
-		"""
-		""" param -> job """
-		jobmap = self.compare_jobs(parameters, filter=filter)
-		""" job -> result """
-		results = self.result(jobs=jobmap.values())
-		""" param -> result [if complete] """
-		return OrderedDict((parval, results[job]) for parval, job in jobmap.items() if results[job] is not None)
+	# @staticmethod
+	# def post_summary(queue, *args, **kwargs):
+	# 	pass
 
 	def get_crash_reason(self, parallel=None, verbosity=0):
 		"""
@@ -700,6 +653,7 @@ class Queue(object):
 
 	def crash_reason(self, parallel=None, verbosity=0, line_len=70, *args, **kwargs):
 		reasons = self.get_crash_reason(parallel=parallel, verbosity=verbosity)
+		if not reasons: return
 		max_name_len = max((len(job.batch_name) + len(job.name)) for job in reasons.keys()) + 1
 		log_template = '{0:' + str(max_name_len) + 's} {1:s}\n'
 		for job, reason in reasons.items():
@@ -730,18 +684,68 @@ class Queue(object):
 				process.communicate()
 				cmd_count += 1
 		self._log('ran command `{1:s}` for {0:d} jobs'.format(cmd_count, cmd), level=1)
-
+	
+	def filter_jobs(self, arg, jobs):
+		"""
+		Filter jobs by pattern or file, for --jobs argument.
+		"""
+		keep_jobs = []
+		if '/' in arg or '\\' in arg:
+			self._log('jobs argument "{0:s}" interpreted as file since it contains / or \\'.format(arg), level=3)
+			assert isfile(arg), '--jobs argument seems to be a file path, but the file does not exist'
+			with open(arg, 'r+') as fh:
+				requested = set(pth.strip() for pth in fh.read().splitlines())
+				if '' in requested: requested.remove('')
+				self._log('filtering jobs by {0:d} names from "{1:s}"'.format(len(requested), arg), level=2)
+				for job in jobs:
+					if job.name in requested:
+						keep_jobs.append(job)
+						requested.remove(job.name)
+				# self.jobs = keep
+				unmatched = requested
+		else:
+			parts = arg.split()
+			self._log('{0:d} jobs arguments interpreted as patterns since they do not contains / or \\'
+				.format(len(parts)), level=3)
+			requested = dict((ptrn, 0) for ptrn in parts)
+			self._log('filtering jobs by {0:d} patterns'.format(len(requested)), level=2)
+			if len(jobs) * len(requested) > 500:
+				warning('many jobs and/or many --jobs filters; this may take a while')
+			for job in jobs:
+				do_keep = False
+				if job.name in requested:
+					# Shortcut without patterns
+					requested[job.name] += 1
+					do_keep = True
+				else:
+					# No literal match, try patterns (slow)
+					for ptrn in requested.keys():
+						if fnmatch(job.name, ptrn):
+							requested[ptrn] += 1
+							do_keep = True
+				if do_keep:
+					keep_jobs.append(job)
+			unmatched = tuple(ptrn for ptrn, cnt in requested.items() if not cnt)
+		if unmatched:
+			raise ValueError('Specifically requested job(s) [{0:s}] was/were not found.'
+				.format(', '.join(unmatched)))
+		return keep_jobs
+	
 	def run_argv(self):
 		"""
 		Analyze sys.argv and run commands based on it.
 		"""
-		def summary(queue = self, *args, **kwargs): return self.summary(queue)
+		def wrap_summary(queue=self, *args, **kwargs):
+			# self.pre_summary(queue, *args, **kwargs)
+			output = self.summary(queue)
+			# self.post_summary(queue, *args, **kwargs)
+			return output
 		def wrap_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd)
 		def wrap_prep_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.PREPARED)
 		def wrap_run_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.RUNNING)
 		def wrap_comp_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.COMPLETED)
 		def wrap_crash_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.CRASHED)
-		parser=ArgumentParser(description='distribute jobs over available nodes',
+		parser = ArgumentParser(description='distribute jobs over available nodes',
 			epilog='actions are executed (largely) in the order they are supplied; some actions may call others where necessary')
 		parser.add_argument('-v', '--verbose', dest='verbosity', action='count', default=0, help='more information (can be used multiple times, -vv)')
 		parser.add_argument('-f', '--force', dest='force', action='store_true', help='force certain mistake-sensitive steps instead of failing with a warning')
@@ -749,21 +753,23 @@ class Queue(object):
 		parser.add_argument('-a', '--availability', dest='availability', action='store_true', help='list all available nodes and their load (cache reload)')
 		parser.add_argument('-d', '--distribute', dest='distribute', action='store_true', help='distribute the jobs over available nodes')
 		parser.add_argument('-l', '--list', dest='actions', action='append_const', const=self.list_jobs, help='show a list of added jobs')
+		parser.add_argument('--json', dest='actions', action='append_const', const=self.json_jobs, help='print job info in fixed-width json format')
 		parser.add_argument('-p', '--prepare', dest='actions', action='append_const', const=self.prepare, help='prepare all the jobs')
 		parser.add_argument('-c', '--calc', dest='actions', action='append_const', const=self.start, help='start calculating one jobs, or see -z/-w/-q')
 		#parser.add_argument('-b', '--keepcalc', dest='actions', action='append_const', const=None, help='like -c, but keeps checking and filling')
 		# parser.add_argument('-z', '--all', dest='all', action='store_true', help='-c will start all jobs')
-		parser.add_argument('-w', '--weight', dest='weight', action='store', type=int, default=None, help='-c will start jobs with total WEIGHT running')
-		parser.add_argument('-q', '--limit', dest='limit', action='store', type=int, default=None, help='-c will add jobs until a total LIMIT running')
+		parser.add_argument('-w', '--weight', dest='weight', action='store', type=int, default=None, help='-c will start jobs with total weight is running')
+		parser.add_argument('-q', '--limit', dest='limit', action='store', type=int, default=None, help='-c will add jobs until a total amount is running')
 		parser.add_argument('-k', '--kill', dest='actions', action='append_const', const=self.kill, help='terminate the calculation of all the running jobs')
 		parser.add_argument('-r', '--remove', dest='actions', action='append_const', const=self.cleanup, help='clean up all the job files')
 		parser.add_argument('-g', '--fix', dest='actions', action='append_const', const=self.fix, help='fix jobs, check cache etc (e.g. after update)')
 		parser.add_argument('-s', '--status', dest='actions', action='append_const', const=self.status, help='show job status')
 		parser.add_argument('-m', '--monitor', dest='actions', action='append_const', const=self.continuous_status, help='show job status every few seconds')
-		parser.add_argument('-x', '--result', dest='actions', action='append_const', const=summary, help='run analysis code to summarize results')
+		parser.add_argument('-x', '--result', dest='actions', action='append_const', const=wrap_summary, help='run analysis code to summarize results')
 		parser.add_argument('-t', '--whyfail', dest='actions', action='append_const', const=self.crash_reason, help ='print a list of failed jobs with the reason why they failed')
 		parser.add_argument('-j', '--serial', dest='parallel', action='store_false', help='make job commands (start, fix, etc) serial (parallel is faster but order is inconsistent)')
-		parser.add_argument('--jobs', dest='jobs', action='store', type=str, help='specify by name the jobs to (re)start or use, separated by whitespace, ?*[] patterns allowed')
+		parser.add_argument('--jobs', dest='jobs', action='store', type=str, help=('if argument contains \ or /, it should be a file containing a job name per line; '
+			'otherwise it is a string specifying by name the jobs to (re)start or use, separated by whitespace, ?*[] patterns allowed'))
 		# parser.add_argument('--cmd', dest='cmd', nargs=1, action='store', type=str, help='run shell command in each of the job directories')
 		parser.add_argument('--cmd', dest='actions', nargs=1, action='append', type=wrap_cmd, help='run a shell command in the directories of each job that has a dir ' + \
 			'($NAME/$BATCH/$STATUS if -s); also prep-cmd, run-cmd, comp-cmd, crash-cmd')
@@ -775,21 +781,16 @@ class Queue(object):
 		""" Note that some other options may be in use by subclass queues. """
 		args = parser.parse_args()
 
+		self.show = args.verbosity + 1
+		self.force, self.restart, self.weight, self.limit = \
+			args.force, args.restart, args.weight, args.limit
+
 		""" Handle only specific jobs by deleting the others (not the most clean perhaps, but good enough for the occasional use). """
 		if args.jobs:
-			requested = dict((ptrn, 0) for ptrn in args.jobs.split())
-			for job in tuple(self.jobs):
-				keep = False
-				for ptrn in requested.keys():
-					if fnmatch(job.name, ptrn):
-						requested[ptrn] += 1
-						keep = True
-				if not keep:
-					self.jobs.remove(job)
-			unmatched = tuple(ptrn for ptrn, cnt in requested.items() if not cnt)
-			if unmatched:
-				parser.error('Specifically requested job(s) [{0:s}] was/were not found.'
-					.format(', '.join(unmatched)))
+			try:
+				self.jobs = self.filter_jobs(args.jobs, self.jobs)
+			except ValueError as err:
+				parser.error(str(err))
 
 		if not args.actions:
 			stderr.write('No action selected. Use one or several flags to control actions.\n\n')
@@ -803,10 +804,6 @@ class Queue(object):
 			else:
 				actions.append(act)
 
-		self.show = args.verbosity + 1
-		self.force, self.restart, self.weight, self.limit = \
-			args.force, args.restart, args.weight, args.limit
-
 		if not actions and not any((args.availability, args.distribute, self.restart, self.weight, self.limit,)):
 			parser.print_help()
 			parser.error('Please provide some action.')
@@ -841,9 +838,9 @@ class Queue(object):
 		return [str(action) for action in actions]
 
 
-def _start_job_on_node(info, **kwargs):
-	job, node = info
+# def _start_job_on_node(info, **kwargs):
+# 	job, node = info
 	# job.cleanup(**kwargs)
-	return job.start(node, **kwargs)
+	# return job.start(node, **kwargs)
 
 
